@@ -23,7 +23,7 @@
  * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#ident "$Id: depmod.c 1.10 Tue, 02 Oct 2001 12:28:38 +1000 kaos $"
+#ident "$Id: depmod.c 1.14 Sun, 11 Nov 2001 16:27:41 +1100 kaos $"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -203,6 +203,19 @@ struct ieee1394_device_id {
 	ElfW(Addr) driver_data;
 };
 
+/*
+ * Device tables for pnpbios devices.
+ * Added Nov. 1, 2001, Andrey Panin <pazke@orbita1.ru>.
+ * See /include/linux/pnp_bios.h
+ */
+
+unsigned int pnpbios_device_id_ver;
+
+struct pnpbios_device_id {
+	char id[8];
+	ElfW(Addr) driver_data;
+};
+
 typedef struct MODULE {
 	char *name;
 	int resolved;
@@ -224,6 +237,8 @@ typedef struct MODULE {
 	int n_parport_device;
 	struct ieee1394_device_id *ieee1394_device;
 	int n_ieee1394_device;
+	struct pnpbios_device_id *pnpbios_device;
+	int n_pnpbios_device;
 } MODULE;
 
 typedef enum SYM_STATUS {
@@ -357,6 +372,10 @@ static void resolve(void)
 		for (resolved = 1, n = 0, sym = mod->undefs.symtab;
 		     n < mod->undefs.n_syms; ++sym, ++n) {
 			SYMBOL *psym = *sym, *find;
+			const char *name;
+			unsigned short hashval;
+			unsigned short hashval_nosuffix;
+			unsigned short length_nosuffix;
 
 			if (psym->status != SYM_UNDEF)
 				continue;
@@ -371,29 +390,70 @@ static void resolve(void)
 			   comparision then there are two hashes to try.
 			 */
 
-			find = symhash[psym->hashval];
+			name = psym->name;
+			hashval = psym->hashval;
+			hashval_nosuffix = psym->hashval_nosuffix;
+			length_nosuffix = psym->length_nosuffix;
+#ifdef	ARCH_ppc64
+		look_again:
+#endif
+			find = symhash[hashval];
 			while (find) {
-				if (strcmp(find->name, psym->name) == 0) {
+				if (strcmp(find->name, name) == 0) {
 					psym->status = SYM_RESOLVED;
 					psym->module = find->module;
 				}
 				find = find->next;
 			}
-			if (psym->status == SYM_UNDEF && psym->length_nosuffix) {
+			if (psym->status == SYM_UNDEF && length_nosuffix) {
 				/* Try again ignoring the suffix */
-				find = symhash[psym->hashval_nosuffix];
+				find = symhash[hashval_nosuffix];
 				while (find) {
-					if (strncmp(find->name, psym->name,
-					    psym->length_nosuffix) == 0 &&
-					    !find->name[psym->length_nosuffix]) {
+					if (strncmp(find->name, name,
+					    length_nosuffix) == 0 &&
+					    !find->name[length_nosuffix]) {
 						psym->status = SYM_RESOLVED;
 						psym->module = find->module;
 					}
 					find = find->next;
 				}
 			}
-			if (psym->status == SYM_UNDEF)
+			if (psym->status == SYM_UNDEF) {
+#ifdef	ARCH_ppc64
+				/* ppc64 is one of those architectures with
+				   function descriptors.  A function is
+				   exported and accessed across object
+				   boundaries via its function descriptor.
+				   The function code symbol happens to be the
+				   function name, prefixed with '.', and a
+				   function call is a branch to the code
+				   symbol.  The linker recognises when a call
+				   crosses object boundaries, and inserts a
+				   stub to call via the function descriptor.
+				   obj_ppc64.c of course does the same thing,
+				   so here we recognise that an undefined code
+				   symbol can be satisfied by the
+				   corresponding function descriptor symbol.
+				*/
+				if (name == psym->name && name[0] == '.') {
+					const char *pt;
+
+					name++;
+					if (length_nosuffix != 0)
+						length_nosuffix--;
+					hashval = 0;
+					for (pt = name; *pt; ++pt) {
+						if (pt == name + length_nosuffix)
+							hashval_nosuffix = hashval;
+						hashval = (hashval << 1) ^ *pt;
+					}
+					hashval %= 2048;
+					hashval_nosuffix %= 2048;
+					goto look_again;
+				}
+#endif	/* ARCH_ppc64 */
 				resolved = 0;
+			}
 		}
 		mod->resolved = resolved;
 	}
@@ -668,6 +728,38 @@ static void extract_parport_device_id(struct obj_file *f, void *image, unsigned
 	}
 }
 
+static void extract_pnpbios_device_id(struct obj_file *f, void *image, unsigned long m_size, MODULE *mod)
+{
+	struct pnpbios_device_id pnpbios_device, *latest;
+	ElfW(Addr) ref_pnpbios, ref_ref_pnpbios;
+	unsigned tgt_long pnpbios_device_size;
+
+	ref_pnpbios = obj_symbol_final_value(f, obj_find_symbol(f, "__module_pnpbios_device_size"));
+	if (!in_range(f, m_size, ref_pnpbios, sizeof(pnpbios_device_size)))
+		return;
+	memcpy(&pnpbios_device_size, (char *)image + ref_pnpbios - f->baseaddr, sizeof(pnpbios_device_size));
+	ref_ref_pnpbios = obj_symbol_final_value(f, obj_find_symbol(f, "__module_pnpbios_device_table"));
+	if (!in_range(f, m_size, ref_ref_pnpbios, sizeof(ref_pnpbios)))
+		return;
+	memcpy(&ref_pnpbios, (char *)image + ref_ref_pnpbios - f->baseaddr, sizeof(ref_pnpbios));
+	extract_version(f, image, m_size, "pnpbios_device", &pnpbios_device_id_ver, 1, 1);
+	if (pnpbios_device_size != sizeof(pnpbios_device)) {
+		error("Unexpected value (%" tgt_long_fmt "d) in '%s' for pnpbios_device_size%s",
+			pnpbios_device_size, f->filename, has_kernel_changed);
+		exit(-1);
+	}
+	while (in_range(f, m_size, ref_pnpbios, pnpbios_device_size)) {
+		memset(&pnpbios_device, 0, sizeof(pnpbios_device));
+		memcpy(&pnpbios_device, (char *)image + ref_pnpbios - f->baseaddr, pnpbios_device_size);
+		ref_pnpbios += pnpbios_device_size;
+		if (!pnpbios_device.id[0])
+			break;
+		mod->pnpbios_device = xrealloc(mod->pnpbios_device, ++(mod->n_pnpbios_device)*sizeof(*(mod->pnpbios_device)));
+		latest = mod->pnpbios_device + mod->n_pnpbios_device-1;
+		*latest = pnpbios_device;
+	}
+}
+
 /*
  *	Extract any ieee1394_device_table from the module.
  *	Note: assumes same machine and arch for depmod and module.
@@ -865,6 +957,13 @@ static int loadobj(const char *objname, int ignore_suffix)
 		mod->resolved = 1;
 	}
 
+#ifdef	ARCH_ppc64
+	if (!ppc64_process_syms (f)) {
+		error("ppc64_process_syms failed\n");
+		return(-1);
+	}
+#endif	/* ARCH_ppc64 */
+
 	/* Do a pseudo relocation to base address 0x1000 (arbitrary).
 	 * All undefined symbols are treated as absolute 0.  This builds
 	 * enough of a module to allow extraction of internal data such
@@ -886,6 +985,7 @@ static int loadobj(const char *objname, int ignore_suffix)
 	extract_usb_device_id(f, image, m_size, mod);
 	extract_parport_device_id(f, image, m_size, mod);
 	extract_ieee1394_device_id(f, image, m_size, mod);
+	extract_pnpbios_device_id(f, image, m_size, mod);
 	free(image);
 
 	obj_free(f);
@@ -987,7 +1087,23 @@ static int addksyms(char *file_syms)
 				if (*p != '?')
 					continue;
 				p = strtok(NULL, " \t\n");
-				if (!strncmp(p, "__kstrtab_", 10))
+				/* Sparc has symbols like '.div' that need to be
+				 * exported.  They strip the '.' and prefix the
+				 * symbol with '__sparc_dot_'.  There is no need
+				 * for 'sparc' in the name, other systems with
+				 * the same problem should use just '__dot_'.
+				 * Support both prefixes, ready for other
+				 * systems.
+				 * */
+				if (!strncmp(p, "__kstrtab___dot_", 15)) {
+					p += 15;
+					*p = '.';
+				}
+				else if (!strncmp(p, "__kstrtab___sparc_dot_", 21)) {
+					p += 21;
+					*p = '.';
+				}
+				else if (!strncmp(p, "__kstrtab_", 10))
 					p += 10;
 				else if (!strncmp(p, "__export_priv_", 14)) {
 					/* Sparc has some weird exported symbols marked
@@ -1091,6 +1207,7 @@ static int prtdepend(char *base_dir, int nflag)
 	FILE *usbmap = stdout;
 	FILE *parportmap = stdout;
 	FILE *ieee1394map = stdout;
+	FILE *pnpbiosmap = stdout;
 	MODULE **tbdep;
 	MODULE *ptmod;
 	int i;
@@ -1105,6 +1222,7 @@ static int prtdepend(char *base_dir, int nflag)
 		usbmap = gen_file_open(gen_file+GEN_USBMAPFILE);
 		parportmap = gen_file_open(gen_file+GEN_PARPORTMAPFILE);
 		ieee1394map = gen_file_open(gen_file+GEN_IEEE1394MAPFILE);
+		pnpbiosmap = gen_file_open(gen_file+GEN_PNPBIOSMAPFILE);
 	}
 
 	skipchars = strlen(base_dir);
@@ -1291,7 +1409,7 @@ static int prtdepend(char *base_dir, int nflag)
 		if (!ptmod->n_ieee1394_device)
 			continue;
 		for (j = 0; j < ptmod->n_ieee1394_device; j++, ieee1394_device++) {
-			fprintf(ieee1394map, "%-20s 0x%08x  0x%06x  0x%06x 0x%06x     0x%06x\n", 
+			fprintf(ieee1394map, "%-20s 0x%08x  0x%06x  0x%06x 0x%06x     0x%06x\n",
 				shortname(ptmod->name),
 				ieee1394_device->match_flags,
 				ieee1394_device->vendor_id,
@@ -1301,6 +1419,20 @@ static int prtdepend(char *base_dir, int nflag)
 		}
 	}
 
+	ptmod = modules;
+	fprintf(pnpbiosmap, "# module             id\n");
+	for (i = 0; i < n_modules; i++, ptmod++) {
+		int j;
+		struct pnpbios_device_id *pnpbios_device = ptmod->pnpbios_device;
+		if (!ptmod->n_pnpbios_device)
+			continue;
+		for (j = 0; j < ptmod->n_pnpbios_device; j++, pnpbios_device++) {
+			fprintf(pnpbiosmap, "%-20s %s\n",
+				shortname(ptmod->name),
+				(char *)pnpbios_device->id);
+		}
+	}
+
 	if (generic_string != stdout)
 		fclose(generic_string);
 	if (pcimap != stdout)
@@ -1313,6 +1445,8 @@ static int prtdepend(char *base_dir, int nflag)
 		fclose(parportmap);
 	if (ieee1394map != stdout)
 		fclose(ieee1394map);
+	if (pnpbiosmap != stdout)
+		fclose(pnpbiosmap);
 	/* Close depfile last, it's timestamp is critical */
 	if (dep != stdout)
 		fclose(dep);
