@@ -20,7 +20,7 @@
    along with this program; if not, write to the Free Software Foundation,
    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
 
-#ident "$Id: obj_ppc.c 1.3 Sat, 22 Sep 2001 20:06:32 +1000 kaos $"
+#ident "$Id: obj_ppc.c 1.4 Wed, 26 Sep 2001 14:50:45 +1000 kaos $"
 
 #include <stddef.h>
 #include <module.h>
@@ -180,8 +180,8 @@ arch_create_got (struct obj_file *f)
   int i, offset;
   struct obj_section *sec, *syms, *strs;
   ElfW(Rela) *rel, *relend;
-  ElfW(Sym) *symtab, *extsym;
-  const char *strtab, *name;
+  ElfW(Sym) *symtab;
+  const char *strtab;
   struct ppc_symbol *intsym;
   struct ppc_plt_entry *pe;
 
@@ -203,12 +203,7 @@ arch_create_got (struct obj_file *f)
 	{
 	  if (ELF32_R_TYPE(rel->r_info) != R_PPC_REL24)
 	    continue;
-	  extsym = &symtab[ELF32_R_SYM(rel->r_info)];
-	  if (extsym->st_name)
-	    name = strtab + extsym->st_name;
-	  else
-	    name = f->sections[extsym->st_shndx]->name;
-	  intsym = (struct ppc_symbol *) obj_find_symbol(f, name);
+	  obj_find_relsym(intsym, f, f, rel, symtab, strtab);
 
 	  for (pe = intsym->plt_entries; pe != NULL; pe = pe->next)
 	    if (pe->addend == rel->r_addend)
